use io;

/* Read instance data. The input files follow the "Augerat" format */
function input() {
    usage = "Usage: localsolver cvrp.lsp inFileName=inputFile "
        + "[solFileName=outputFile] [lsTimeLimit=timeLimit] [nbTrucks=value]";

    if (inFileName == nil) throw usage;
    
    readInputCvrp();
    
    // The number of trucks is usually given in the name of the file
    // nbTrucks can also be given in command line
    if (nbTrucks == nil) nbTrucks = getNbTrucks();

    computeDistanceMatrix();
}
 
/* Declare the optimization model */
function model() {
    // Sequence of customers visited by each truck
    customersSequences[k in 0...nbTrucks] <- list(nbCustomers);

    // All customers must be visited by exactly one truck
    constraint partition[k in 0...nbTrucks](customersSequences[k]);

    for [k in 0...nbTrucks] {
        local sequence <- customersSequences[k];
        local c <- count(sequence);

        // A truck is used if it visits at least one customer
        trucksUsed[k] <- c > 0;

        // The quantity needed in each route must not exceed the truck capacity
        routeQuantity <- sum(sequence, j => demands[j]);
        constraint routeQuantity <= truckCapacity;

        // Distance traveled by truck k
        routeDistances[k] <- sum(1...c, i => distanceMatrix[sequence[i - 1]][sequence[i]])
                + (c > 0 ? (distanceDepot[sequence[0]] + distanceDepot[sequence[c - 1]]) : 0);
    }

    // Total number of trucks used
    nbTrucksUsed <- sum[k in 0...nbTrucks](trucksUsed[k]);

    // Total distance traveled
    totalDistance <- sum[k in 0...nbTrucks](routeDistances[k]);

    // Objective: minimize the number of trucks used, then minimize the distance traveled
    minimize nbTrucksUsed;
    minimize totalDistance;
}

/* Parametrize the solver */
function param() { 
    if (lsTimeLimit == nil) lsTimeLimit = 20;
}

/* Write the solution in a file with the following format:
 * - number of trucks used and total distance
 * - for each truck the customers visited (omitting the start/end at the depot) */
function output() {
    if (solFileName == nil) return;
    local outfile = io.openWrite(solFileName);

    outfile.println(nbTrucksUsed.value, " ", totalDistance.value);
    for [k in 0...nbTrucks] {
        if (trucksUsed[k].value != 1) continue;
        // Values in sequence are in 0...nbCustomers.
        // +2 is to put it back in 2...nbCustomers+2
        // as in the data files (1 being the depot)
        for [customer in customersSequences[k].value]
            outfile.print(customer + 2, " ");
        outfile.println();
    }
}


function readInputCvrp() {
    local inFile = io.openRead(inFileName);
    local nbNodes = 0;
    while (true) {
        local str = inFile.readString();
        if (str.startsWith("DIMENSION")) {
            if (!str.endsWith(":")) str = inFile.readString();
            nbNodes = inFile.readInt();
            nbCustomers = nbNodes - 1;
        } else if ((str.startsWith("CAPACITY"))) {
            if (!str.endsWith(":")) str = inFile.readString();
            truckCapacity = inFile.readInt();
        } else if ((str.startsWith("EDGE_WEIGHT_TYPE"))) {
            if (!str.endsWith(":")) str = inFile.readString();
            local weightType = inFile.readString();
            if (weightType != "EUC_2D")
                throw "Edge Weight Type " + weightType + " is not supported (only EUC_2D)";
        } else if (str.startsWith("NODE_COORD_SECTION")) {
            break;
        } else {
            local dump = inFile.readln();
        }
    }

    // -2 because original customer indices are in 2..nbNodes
    for [n in 0...nbNodes] {
        local node_id = inFile.readInt();
        if (n + 1 != node_id) throw "Unexpected index";
        if (node_id == 1) {
            depotX = round(inFile.readDouble());
            depotY = round(inFile.readDouble());
        } else {
            customersX[node_id - 2] = round(inFile.readDouble());
            customersY[node_id - 2] = round(inFile.readDouble());
        }
    }

    dump = inFile.readln();
    if (!dump.startsWith("DEMAND_SECTION")) throw "Expected keyword DEMAND_SECTION";
    for [n in 1..nbNodes] {
        if (n != inFile.readInt()) throw "Unexpected index";
        local demand = inFile.readInt();
        if (n == 1) {
            if (demand != 0) throw "Demand for depot should be O";
        } else {
            demands[n - 2] = demand; // -2 because original customer indices are in 2..nbNodes
        }
    }

    dump = inFile.readln();
    if (!dump.startsWith("DEPOT_SECTION")) throw "Expected keyword DEPOT_SECTION";
    local depotId = inFile.readInt();
    if (depotId != 1) throw "Depot id is supposed to be 1";
    local endOfDepotSection = inFile.readInt();
    if (endOfDepotSection != -1) throw "Expecting only one depot, more than one found";
}

/* Compute the distance matrix */
function computeDistanceMatrix() {
    for [i in 0...nbCustomers] {
        distanceMatrix[i][i] = 0;
        for [j in i+1...nbCustomers] {
            local localDistance = computeDist(customersX[i], customersX[j], customersY[i], customersY[j]);
            distanceMatrix[j][i] = localDistance;
            distanceMatrix[i][j] = localDistance;
        }
    }

    for [i in 0...nbCustomers] {
        local localDistance = computeDist(depotX, customersX[i], depotY, customersY[i]);
        distanceDepot[i] = localDistance;
    }
}

function computeDist(xi, xj, yi, yj) {
    local exactDist = sqrt(pow((xi - xj), 2) + pow((yi - yj), 2));
    return round(exactDist);
}

function getNbTrucks() {
    local splitted = inFileName.split("-k");
    if (count(splitted) >= 2) {
        local numvrp = splitted[count(splitted) - 1];
        splitted = numvrp.split(".");
        if (count(splitted) == 2) return splitted[0].toInt();
    } else {
        println("Error: nbTrucks could not be read from the file name.
                Enter it from the command line");
        throw usage;
    }
}
